/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.api;


import java.util.Set;

import org.crazyjavahacking.astvisualizer.api.model.EdgeModel;
import org.crazyjavahacking.astvisualizer.api.model.NodeModel;


/**
 * Customized Abstract Syntax Tree. The AST is immutable data structure. AST extends NodeModel Iterable using
 * <b>Breadth First</b> algorithm.
 *
 * @author Martin
 *
 * @since 1.0.3
 */
public interface CustomizedAST extends Iterable<NodeModel> {

    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Returns incoming edge of given node. For root node which has no incoming edge, <b>EdgeModel.NO_INCOMING_EDGE</b>
     * will be returned.
     *
     * @param node node
     *
     * @return incoming edge or EdgeModel.NO_INCOMING_EDGE
     */
    EdgeModel getIncomingEdgeOfNode(NodeModel node);

    /**
     * Returns <b>immutable set</b> of outgoing edges of given node. For leaf nodes <b>empty set</b> will be returned.
     *
     * @param node node
     *
     * @return immutable set of outgoing edges or empty set
     */
	Set<EdgeModel> getOutgoingEdgesOfNode(NodeModel node);

    /**
     * Returns source node of given edge.
     *
     * @param edge edge
     *
     * @return source node
     */
	NodeModel getSourceNodeOfEdge(EdgeModel edge);

    /**
     * Returns target node of given edge.
     *
     * @param edge edge
     *
     * @return target node
     */
	NodeModel getTargetNodeOfEdge(EdgeModel edge);

    /**
     * Returns <b>immutable set</b> of all nodes.
     *
     * @return immutable set of nodes
     */
    Set<NodeModel> getAllNodes();

    /**
     * Returns <b>immutable set</b> of all edges.
     *
     * @return immutable set of edges
     */
    Set<EdgeModel> getAllEdges();

    /**
     * Returns textual representation of this customized AST.
     *
     * @return textual representation
     *
     * @since 1.0.4
     */
    String getTextualRepresentation();
}
