/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.impl.javac;


import java.util.Collections;
import java.util.Iterator;
import java.util.Set;

import org.crazyjavahacking.astvisualizer.api.CustomizedAST;
import org.crazyjavahacking.astvisualizer.api.model.EdgeModel;
import org.crazyjavahacking.astvisualizer.api.model.NodeModel;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.SimpleDirectedGraph;
import org.jgrapht.traverse.BreadthFirstIterator;


/**
 * Represents modifiable AST.
 *
 * @author Martin
 */
final class ModifiableAST implements CustomizedAST {

    //-----------------------------------------------------------------------------------------------------------------
    // Members.
    //-----------------------------------------------------------------------------------------------------------------
    private final DirectedGraph<NodeModel, EdgeModel> directedGraph =
            new SimpleDirectedGraph<NodeModel, EdgeModel>(EdgeModel.class);
    private final String textualRepresentation;

    private NodeModel lastAddedNode = null;


    //-----------------------------------------------------------------------------------------------------------------
    // Constructors.
    //-----------------------------------------------------------------------------------------------------------------
    public ModifiableAST(String textualRepresentation) {
        this.textualRepresentation = textualRepresentation;
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Adds node to customized AST.
     *
     * @param node node
     */
    public void addNode(NodeModel node) {
        directedGraph.addVertex(node);

        lastAddedNode = node;
    }

    /**
     * Adds edge to customized AST. Given nodes must be first added using <code>addNode(NodeModel)</code> method.
     *
     * @param startNode start node
     * @param endNode   end node
     * @param edge      edge
     */
    public void addEdge(NodeModel startNode, NodeModel endNode, EdgeModel edge) {
        directedGraph.addEdge(startNode, endNode, edge);
    }

    /**
     * Returns number of nodes.
     *
     * @return number of nodes
     */
    public int getNumberOfNodes() {
        return directedGraph.vertexSet().size();
    }

    /**
     * Returns last added node.
     *
     * @return last added node
     */
    public NodeModel getLastAddedNode() {
        return lastAddedNode;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public EdgeModel getIncomingEdgeOfNode(NodeModel node) {
        Set<EdgeModel> incomingEdges = directedGraph.incomingEdgesOf(node);

        switch (incomingEdges.size()) {
            case 0:
                return EdgeModel.NO_INCOMING_EDGE;

            case 1:
                return incomingEdges.iterator().next();

            default:
                throw new IllegalStateException("Given node has multiple incoming edges! : " + node);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<EdgeModel> getOutgoingEdgesOfNode(NodeModel node) {
        return Collections.unmodifiableSet(directedGraph.outgoingEdgesOf(node));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public NodeModel getSourceNodeOfEdge(EdgeModel edge) {
        return directedGraph.getEdgeSource(edge);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public NodeModel getTargetNodeOfEdge(EdgeModel edge) {
        return directedGraph.getEdgeTarget(edge);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Iterator<NodeModel> iterator() {
        return new BreadthFirstIterator<NodeModel, EdgeModel>(directedGraph);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<NodeModel> getAllNodes() {
        return Collections.unmodifiableSet(directedGraph.vertexSet());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<EdgeModel> getAllEdges() {
        return Collections.unmodifiableSet(directedGraph.edgeSet());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getTextualRepresentation() {
        return textualRepresentation;
    }
}
